import time
import hashlib
import re
import itertools
import operator
import pycurl

# Python 2 and 3
import sys
if sys.version_info >= (3, 0):
    from urllib.parse import urlparse
else:
    from urlparse import urlparse

from threading import Lock
from collections import namedtuple
from collections import defaultdict

from .filter import FuzzResFilter
from .externals.reqresp import Request, Response
from .exception import FuzzExceptBadAPI, FuzzExceptBadOptions, FuzzExceptInternalError, FuzzException
from .facade import Facade, ERROR_CODE
from .mixins import FuzzRequestUrlMixing, FuzzRequestSoupMixing

from .utils import python2_3_convert_to_unicode, python2_3_convert_from_unicode
from .utils import MyCounter
from .utils import rgetattr
from .utils import DotDict

auth_header = namedtuple("auth_header", "method credentials")


class headers(object):
    class header(DotDict):
        def __str__(self):
            return "\n".join(["{}: {}".format(k, v) for k, v in self.items()])

    def __init__(self, req):
        self._req = req

    @property
    def response(self):
        return headers.header(self._req.response.getHeaders()) if self._req.response else {}

    @property
    def request(self):
        return headers.header([x.split(": ", 1) for x in self._req.getHeaders()])

    @request.setter
    def request(self, values_dict):
        self._req._headers.update(values_dict)
        if "Content-Type" in values_dict:
            self._req.ContentType = values_dict['Content-Type']

    @property
    def all(self):
        return headers.header(self.request + self.response)


class cookies(object):
    class cookie(DotDict):
        def __str__(self):
            return "\n".join(["{}={}".format(k, v) for k, v in self.items()])

    def __init__(self, req):
        self._req = req

    @property
    def response(self):
        if self._req.response:
            c = self._req.response.getCookie().split("; ")
            if c[0]:
                return cookies.cookie([[x[0], x[2]] for x in [x.partition("=") for x in c]])

        return cookies.cookie({})

    @property
    def request(self):
        if 'Cookie' in self._req._headers:
            c = self._req._headers['Cookie'].split("; ")
            if c[0]:
                return cookies.cookie([[x[0], x[2]] for x in [x.partition("=") for x in c]])

        return cookies.cookie({})

    @request.setter
    def request(self, values):
        self._req._headers["Cookie"] = "; ".join(values)

    @property
    def all(self):
        return cookies.cookie(self.request + self.response)


class params(object):
    class param(DotDict):
        def __str__(self):
            return "\n".join(["{}={}".format(k, v) for k, v in self.items()])

    def __init__(self, req):
        self._req = req

    @property
    def get(self):
        return params.param([(x.name, x.value) for x in self._req.getGETVars()])

    @get.setter
    def get(self, values):
        if isinstance(values, dict):
            for key, value in values.items():
                self._req.setVariableGET(key, str(value))
        else:
            raise FuzzExceptBadAPI("GET Parameters must be specified as a dictionary")

    @property
    def post(self):
        return params.param([(x.name, x.value) for x in self._req.getPOSTVars()])

    @post.setter
    def post(self, pp):
        if isinstance(pp, dict):
            for key, value in pp.items():
                self._req.setVariablePOST(key, str(value) if value is not None else value)

            self._req._non_parsed_post = self._req._variablesPOST.urlEncoded()

        elif isinstance(pp, str):
            self._req.setPostData(pp)

    @property
    def raw_post(self):
        return self._req._non_parsed_post

    @property
    def all(self):
        return params.param(self.get + self.post)

    @all.setter
    def all(self, values):
        self.get = values
        self.post = values


class FuzzRequest(FuzzRequestUrlMixing, FuzzRequestSoupMixing):
    def __init__(self):
        self._request = Request()

        self._proxy = None
        self._allvars = None
        self.wf_fuzz_methods = None
        self.wf_retries = 0
        self.wf_ip = None

        self.headers.request = {"User-Agent": Facade().sett.get("connection", "user-agent")}

    # methods for accessing HTTP requests information consistenly accross the codebase

    def __str__(self):
        return self._request.getAll()

    @property
    def raw_content(self):
        if self._request.response:
            return self._request.response.getAll()

        return ""

    @property
    def headers(self):
        return headers(self._request)

    @property
    def params(self):
        return params(self._request)

    @property
    def cookies(self):
        return cookies(self._request)

    @property
    def method(self):
        return self._request.method

    @method.setter
    def method(self, method):
        self._request.method = method

    @property
    def scheme(self):
        return self._request.schema

    @scheme.setter
    def scheme(self, s):
        self._request.schema = s

    @property
    def host(self):
        return self._request.getHost()

    @property
    def path(self):
        return self._request.path

    @property
    def redirect_url(self):
        return self._request.completeUrl

    @property
    def url(self):
        return self._request.finalUrl

    @url.setter
    def url(self, u):
        # urlparse goes wrong with IP:port without scheme (https://bugs.python.org/issue754016)
        if not u.startswith("FUZ") and (urlparse(u).netloc == "" or urlparse(u).scheme == ""):
            u = "http://" + u

        if urlparse(u).path == "":
            u += '/'

        if Facade().sett.get("general", "encode_space") == "1":
            u = u.replace(" ", "%20")

        self._request.setUrl(u)
        if self.scheme.startswith("fuz") and self.scheme.endswith("z"):
            # avoid FUZZ to become fuzz
            self.scheme = self.scheme.upper()

    @property
    def content(self):
        return self._request.response.getContent() if self._request.response else ""

    @property
    def code(self):
        return self._request.response.code if self._request.response else 0

    @code.setter
    def code(self, c):
        self._request.response.code = int(c)

    @property
    def auth(self):
        m, up = self._request.getAuth()
        return auth_header(m, up)

    @auth.setter
    def auth(self, ah):
        method, credentials = ah
        self._request.setAuth(method, credentials)

    @property
    def follow(self):
        return self._request.followLocation

    @follow.setter
    def follow(self, f):
        self._request.setFollowLocation(f)

    @property
    def reqtime(self):
        return self._request.totaltime

    @reqtime.setter
    def reqtime(self, t):
        self._request.totaltime = t

    # Info extra that wfuzz needs within an HTTP request
    @property
    def wf_allvars_set(self):
        if self.wf_allvars == "allvars":
            return self.params.get
        elif self.wf_allvars == "allpost":
            return self.params.post
        elif self.wf_allvars == "allheaders":
            return self.headers.request
        else:
            raise FuzzExceptBadOptions("Unknown variable set: " + self.wf_allvars)

    @wf_allvars_set.setter
    def wf_allvars_set(self, varset):
        try:
            if self.wf_allvars == "allvars":
                self.params.get = varset
            elif self.wf_allvars == "allpost":
                self.params.post = varset
            elif self.wf_allvars == "allheaders":
                self._request.headers.request = varset
            else:
                raise FuzzExceptBadOptions("Unknown variable set: " + self.wf_allvars)
        except TypeError:
            raise FuzzExceptBadOptions("It is not possible to use all fuzzing with duplicated parameters.")

    @property
    def wf_allvars(self):
        return self._allvars

    @wf_allvars.setter
    def wf_allvars(self, bl):
        if bl is not None and bl not in ['allvars', 'allpost', 'allheaders']:
            raise FuzzExceptBadOptions("Incorrect all parameters brute forcing type specified, correct values are allvars, allpost or allheaders.")

        self._allvars = bl

    @property
    def wf_proxy(self):
        return self._proxy

    @wf_proxy.setter
    def wf_proxy(self, l):
        if l:
            prox, ptype = l
            self._request.setProxy("%s" % prox, ptype if ptype else "HTML")
        self._proxy = l

    # methods wfuzz needs to perform HTTP requests (this might change in the future).

    def perform(self):
        res = FuzzResult(self, track_id=False)
        return Facade().http_pool.perform(res)

    def to_http_object(self, c):
        pycurl_c = Request.to_pycurl_object(c, self._request)

        if self.wf_ip:
            pycurl_c.setopt(pycurl.CONNECT_TO, ["::{}:{}".format(self.wf_ip['ip'], self.wf_ip['port'])])

        return pycurl_c

    def from_http_object(self, c, bh, bb):
        raw_header = python2_3_convert_from_unicode(bh.decode("utf-8", errors='surrogateescape'))
        return self._request.response_from_conn_object(c, raw_header, bb)

    def update_from_raw_http(self, raw, scheme, raw_response=None, raw_content=None):
        self._request.parseRequest(raw, scheme)

        # Parse request sets postdata = '' when there's POST request without data
        if self.method == "POST" and self.params.raw_post is None:
            self.params.post = ''

        if raw_response:
            rp = Response()
            if not isinstance(raw_response, str):
                raw_response = python2_3_convert_from_unicode(raw_response.decode("utf-8", errors='surrogateescape'))
            rp.parseResponse(raw_response, raw_content)
            self._request.response = rp

        return self._request

    def to_cache_key(self):
        key = self._request.urlWithoutVariables

        dicc = {'g{}'.format(key): True for key in self.params.get.keys()}
        dicc.update({'p{}'.format(key): True for key in self.params.post.keys()})

        # take URL parameters into consideration
        url_params = list(dicc.keys())
        url_params.sort()
        key += "-" + "-".join(url_params)

        return key

    # methods wfuzz needs for substituing payloads and building dictionaries

    def update_from_options(self, options):
        if options["url"] != "FUZZ":
            self.url = options["url"]

        # headers must be parsed first as they might affect how reqresp parases other params
        self.headers.request = dict(options['headers'])

        if options['auth'][0] is not None:
            self.auth = (options['auth'][0], options['auth'][1])

        if options['follow']:
            self.follow = options['follow']

        if options['postdata'] is not None:
            self.params.post = options['postdata']

        if options['connect_to_ip']:
            self.wf_ip = options['connect_to_ip']

        if options['method']:
            self.method = options['method']
            self.wf_fuzz_methods = options['method']

        if options['cookie']:
            self.cookies.request = options['cookie']

        if options['allvars']:
            self.wf_allvars = options['allvars']

    def from_copy(self):
        newreq = FuzzRequest()

        newreq.wf_proxy = self.wf_proxy
        newreq.wf_allvars = self.wf_allvars
        newreq.wf_fuzz_methods = self.wf_fuzz_methods
        newreq.wf_ip = self.wf_ip

        newreq.headers.request = self.headers.request
        newreq.params.post = self.params.raw_post

        newreq.follow = self.follow
        newreq.auth = self.auth
        newreq.url = self.url
        newreq.reqtime = self.reqtime
        newreq.scheme = self.scheme
        newreq.method = self.wf_fuzz_methods if self.wf_fuzz_methods else self.method

        return newreq


class FuzzResultFactory:
    @staticmethod
    def replace_fuzz_word(text, fuzz_word, payload):
        marker_regex = re.compile(r"(%s)(?:\[(.*?)\])?" % (fuzz_word,), re.MULTILINE | re.DOTALL)

        for fuzz_word, field in marker_regex.findall(text):
            if field:
                marker_regex = re.compile(r"(%s)(?:\[(.*?)\])?" % (fuzz_word,), re.MULTILINE | re.DOTALL)
                fields_array = []

                for fuzz_word, field in marker_regex.findall(text):
                    if not field:
                        raise FuzzExceptBadOptions("You must specify a field when using a payload containing a full fuzz request, ie. FUZZ[url], or use FUZZ only to repeat the same request.")

                    try:
                        subs = str(rgetattr(payload, field))
                    except AttributeError:
                        raise FuzzExceptBadOptions("A FUZZ[field] expression must be used with a fuzzresult payload not a string.")

                    text = text.replace("%s[%s]" % (fuzz_word, field), subs)
                    fields_array.append(field)

                return (text, fields_array)
            else:
                try:
                    return (text.replace(fuzz_word, payload), [None])
                except TypeError:
                    raise FuzzExceptBadOptions("Tried to replace {} with a whole fuzzresult payload.".format(fuzz_word))

    @staticmethod
    def from_seed(seed, payload, seed_options):
        newres = seed.from_soft_copy()

        rawReq = str(newres.history)
        rawUrl = newres.history.redirect_url
        scheme = newres.history.scheme
        auth_method, userpass = newres.history.auth

        for payload_pos, payload_content in enumerate(payload, start=1):
            fuzz_word = "FUZ" + str(payload_pos) + "Z" if payload_pos > 1 else "FUZZ"

            fuzz_values_array = []

            # substitute entire seed when using a request payload generator without specifying field
            if fuzz_word == "FUZZ" and seed_options["seed_payload"] and isinstance(payload_content, FuzzResult):
                # new seed
                newres = payload_content.from_soft_copy()
                newres.payload = []

                fuzz_values_array.append(None)

                newres.history.update_from_options(seed_options)
                newres.update_from_options(seed_options)
                rawReq = str(newres.history)
                rawUrl = newres.history.redirect_url
                scheme = newres.history.scheme
                auth_method, userpass = newres.history.auth

            desc = []

            if auth_method and (userpass.count(fuzz_word)):
                userpass, desc = FuzzResultFactory.replace_fuzz_word(userpass, fuzz_word, payload_content)
            if newres.history.redirect_url.count(fuzz_word):
                rawUrl, desc = FuzzResultFactory.replace_fuzz_word(rawUrl, fuzz_word, payload_content)
            if rawReq.count(fuzz_word):
                rawReq, desc = FuzzResultFactory.replace_fuzz_word(rawReq, fuzz_word, payload_content)

            if scheme.count(fuzz_word):
                scheme, desc = FuzzResultFactory.replace_fuzz_word(scheme, fuzz_word, payload_content)

            if desc:
                fuzz_values_array += desc

            newres.payload.append(FuzzPayload(payload_content, fuzz_values_array))

        newres.history.update_from_raw_http(rawReq, scheme)
        newres.history.url = rawUrl
        if auth_method != 'None':
            newres.history.auth = (auth_method, userpass)

        newres.type = FuzzResult.result

        return newres

    @staticmethod
    def from_baseline(fuzzresult, options):
        scheme = fuzzresult.history.scheme
        rawReq = str(fuzzresult.history)
        auth_method, userpass = fuzzresult.history.auth

        # get the baseline payload ordered by fuzz number and only one value per same fuzz keyword.
        b1 = dict([matchgroup.groups() for matchgroup in re.finditer(r"FUZ(\d*)Z(?:\[.*?\])?(?:{(.*?)})?", rawReq, re.MULTILINE | re.DOTALL)])
        b2 = dict([matchgroup.groups() for matchgroup in re.finditer(r"FUZ(\d*)Z(?:\[.*?\])?(?:{(.*?)})?", userpass, re.MULTILINE | re.DOTALL)])
        baseline_control = dict(list(b1.items()) + list(b2.items()))
        baseline_payload = [x[1] for x in sorted(list(baseline_control.items()), key=operator.itemgetter(0))]

        # if there is no marker, there is no baseline request
        if not [x for x in baseline_payload if x is not None]:
            return None

        # remove baseline marker from seed request
        for i in baseline_payload:
            if not i:
                raise FuzzExceptBadOptions("You must supply a baseline value for all the FUZZ words.")
            rawReq = rawReq.replace("{" + i + "}", '')

            if fuzzresult.history.wf_fuzz_methods:
                fuzzresult.history.wf_fuzz_methods = fuzzresult.history.wf_fuzz_methods.replace("{" + i + "}", '')

            if auth_method:
                userpass = userpass.replace("{" + i + "}", '')

        # re-parse seed without baseline markers
        fuzzresult.history.update_from_raw_http(rawReq, scheme)
        if auth_method:
            fuzzresult.history.auth = (auth_method, userpass)

        # create baseline request from seed
        baseline_res = fuzzresult.from_soft_copy()

        # remove field markers from baseline
        marker_regex = re.compile(r"(FUZ\d*Z)\[(.*?)\]", re.DOTALL)
        results = marker_regex.findall(rawReq)
        if results:
            for fw, f in results:
                rawReq = rawReq.replace("%s[%s]" % (fw, f), fw)

                if fuzzresult.history.wf_fuzz_methods:
                    fuzzresult.history.wf_fuzz_methods = fuzzresult.history.wf_fuzz_methods.replace("{" + i + "}", '')

                if auth_method:
                    userpass = userpass.replace("{" + i + "}", '')

            baseline_res.history.update_from_raw_http(rawReq, scheme)

        baseline_res = FuzzResultFactory.from_seed(baseline_res, baseline_payload, options)
        baseline_res.is_baseline = True

        return baseline_res

    @staticmethod
    def from_all_fuzz_request(seed, payload):
        # only a fuzz payload is allowed using this technique
        if len(payload) > 1:
            raise FuzzExceptBadOptions("Only one payload is allowed when fuzzing all parameters!")

        for var_name in seed.history.wf_allvars_set.keys():
            payload_content = payload[0]
            fuzzres = seed.from_soft_copy()
            fuzzres.payload.append(FuzzPayload(payload_content, [None]))

            fuzzres.history.wf_allvars_set = {var_name: payload_content}

            yield fuzzres

    @staticmethod
    def from_options(options):
        fr = FuzzRequest()

        fr.url = options['url']
        fr.wf_fuzz_methods = options['method']
        fr.update_from_options(options)

        fuzz_res = FuzzResult(fr)
        fuzz_res.update_from_options(options)

        return fuzz_res


class FuzzStats:
    def __init__(self):
        self.mutex = Lock()

        self.url = ""
        self.seed = None

        self.total_req = 0
        self.pending_fuzz = MyCounter()
        self.pending_seeds = MyCounter()
        self.processed = MyCounter()
        self.backfeed = MyCounter()
        self.filtered = MyCounter()

        self.totaltime = 0
        self.__starttime = 0

        self._cancelled = False

    @staticmethod
    def from_requestGenerator(rg):
        tmp_stats = FuzzStats()

        tmp_stats.url = rg.seed.history.redirect_url
        tmp_stats.total_req = rg.count()
        tmp_stats.seed = rg.seed

        return tmp_stats

    def get_stats(self):
        return {
            "url": self.url,
            "total": self.total_req,

            "backfed": self.backfeed(),
            "Processed": self.processed(),
            "Pending": self.pending_fuzz(),
            "filtered": self.filtered(),

            "Pending_seeds": self.pending_seeds(),

            "totaltime": self._totaltime,
        }

    def mark_start(self):
        with self.mutex:
            self.__starttime = time.time()

    def mark_end(self):
        with self.mutex:
            self.totaltime = time.time() - self.__starttime

    @property
    def cancelled(self):
        with self.mutex:
            return self._cancelled

    @cancelled.setter
    def cancelled(self, v):
        with self.mutex:
            self._cancelled = v

    def __str__(self):
        string = ""

        string += "Total time: %s\n" % str(self.totaltime)[:8]

        if self.backfeed() > 0:
            string += "Processed Requests: %s (%d + %d)\n" % (str(self.processed())[:8], (self.processed() - self.backfeed()), self.backfeed())
        else:
            string += "Processed Requests: %s\n" % (str(self.processed())[:8])
        string += "Filtered Requests: %s\n" % (str(self.filtered())[:8])
        string += "Requests/sec.: %s\n" % str(self.processed() / self.totaltime if self.totaltime > 0 else 0)[:8]

        return string

    def update(self, fuzzstats2):
        self.url = fuzzstats2.url
        self.total_req += fuzzstats2.total_req
        self.totaltime += fuzzstats2.totaltime

        self.backfeed._operation(fuzzstats2.backfeed())
        self.processed._operation(fuzzstats2.processed())
        self.pending_fuzz._operation(fuzzstats2.pending_fuzz())
        self.filtered._operation(fuzzstats2.filtered())
        self.pending_seeds._operation(fuzzstats2.pending_seeds())


class FuzzPayload():
    def __init__(self, content, fields):
        self.content = content
        self.fields = fields

    def description(self, default):
        ret_str_values = []
        for fuzz_value in self.fields:
            if fuzz_value is None and isinstance(self.content, FuzzResult):
                ret_str_values.append(default)
            elif fuzz_value is not None and isinstance(self.content, FuzzResult):
                ret_str_values.append(str(rgetattr(self.content, fuzz_value)))
            elif fuzz_value is None:
                ret_str_values.append(self.content)
            else:
                ret_str_values.append(fuzz_value)

        return " - ".join(ret_str_values)

    def __str__(self):
        return "content: {} fields: {}".format(self.content, self.fields)


class FuzzResult:
    seed, backfeed, result, error, startseed, endseed, cancel, discarded = list(range(8))
    newid = itertools.count(0)

    def __init__(self, history=None, exception=None, track_id=True):
        self.history = history

        self.type = None
        self.exception = exception
        self.is_baseline = False
        self.rlevel = 1
        self.nres = next(FuzzResult.newid) if track_id else 0

        self.chars = 0
        self.lines = 0
        self.words = 0
        self.md5 = ""

        self.update()

        self.plugins_res = []
        self.plugins_backfeed = []

        self.payload = []

        self._description = None
        self._show_field = False

    @property
    def plugins(self):
        dic = defaultdict(list)

        for pl in self.plugins_res:
            dic[pl.source].append(pl.issue)

        return dic

    def update(self, exception=None):
        self.type = FuzzResult.result

        if exception:
            self.exception = exception

        if self.history and self.history.content:
            m = hashlib.md5()
            m.update(python2_3_convert_to_unicode(self.history.content))
            self.md5 = m.hexdigest()

            self.chars = len(self.history.content)
            self.lines = self.history.content.count("\n")
            self.words = len(re.findall(r"\S+", self.history.content))

        return self

    def __str__(self):
        if self.type == FuzzResult.result:
            res = "%05d:  C=%03d   %4d L\t   %5d W\t  %5d Ch\t  \"%s\"" % (self.nres, self.code, self.lines, self.words, self.chars, self.description)
            for i in self.plugins_res:
                res += "\n  |_ %s" % i.issue

            return res
        else:
            return "Control result, type: %s" % ("seed", "backfeed", "result", "error", "startseed", "endseed", "cancel", "discarded")[self.type]

    def _payload_description(self):
        if not self.payload:
            return self.url

        payl_descriptions = [payload.description(self.url) for payload in self.payload]
        ret_str = ' - '.join([p_des for p_des in payl_descriptions if p_des])

        return ret_str

    @property
    def description(self):
        ret_str = ""

        if self._show_field is True:
            ret_str = self.eval(self._description)
        elif self._show_field is False and self._description is not None:
            ret_str = "{} | {}".format(self._payload_description(), self.eval(self._description))
        else:
            ret_str = self._payload_description()

        if self.exception:
            return ret_str + "! " + str(self.exception)

        return ret_str

    def eval(self, expr):
        return FuzzResFilter(filter_string=expr).is_visible(self)

    # parameters in common with fuzzrequest
    @property
    def content(self):
        return self.history.content if self.history else ""

    @property
    def url(self):
        return self.history.url if self.history else ""

    @property
    def code(self):
        if self.history and self.history.code >= 0 and not self.exception:
            return int(self.history.code)
        # elif not self.history.code:
            # return 0
        else:
            return ERROR_CODE

    @property
    def timer(self):
        return self.history.reqtime if self.history and self.history.reqtime else 0

    # factory methods

    def to_new_seed(self):
        seed = self.from_soft_copy(False)

        if seed.type == FuzzResult.error:
            raise FuzzExceptInternalError("A new seed cannot be created with a Fuzz item representing an error.")

        seed.history.url = self.history.recursive_url
        seed.rlevel += 1
        seed.type = FuzzResult.seed

        return seed

    def from_soft_copy(self, track_id=True):
        fr = FuzzResult(self.history.from_copy(), track_id=track_id)

        fr.exception = self.exception
        fr.is_baseline = self.is_baseline
        fr.type = self.type
        fr.rlevel = self.rlevel
        fr.payload = list(self.payload)
        fr._description = self._description
        fr._show_field = self._show_field

        return fr

    def update_from_options(self, options):
        self._description = options['description']
        self._show_field = options['show_field']

    @staticmethod
    def to_new_exception(exception):
        fr = FuzzResult(exception=exception, track_id=False)
        fr.type = FuzzResult.error

        return fr

    @staticmethod
    def to_new_signal(signal):
        fr = FuzzResult(track_id=False)
        fr.type = signal

        return fr

    def to_new_url(self, url):
        fr = self.from_soft_copy()
        fr.history.url = str(url)
        fr.rlevel = self.rlevel + 1
        fr.type = FuzzResult.backfeed
        fr.is_baseline = False

        return fr

    def __lt__(self, other):
        return self.nres < other.nres

    def __le__(self, other):
        return self.nres <= other.nres

    def __gt__(self, other):
        return self.nres > other.nres

    def __ge__(self, other):
        return self.nres >= other.nres

    def __eq__(self, other):
        return self.nres == other.nres

    def __ne__(self, other):
        return self.nres != other.nres


class PluginItem:
    undefined, result, backfeed = list(range(3))

    def __init__(self, ptype):
        self.source = ""
        self.plugintype = ptype


class PluginResult(PluginItem):
    def __init__(self):
        PluginItem.__init__(self, PluginItem.result)

        self.issue = ""


class PluginRequest(PluginItem):
    def __init__(self):
        PluginItem.__init__(self, PluginItem.backfeed)

        self.fuzzitem = None

    @staticmethod
    def from_fuzzRes(res, url, source):
        plreq = PluginRequest()
        plreq.source = source
        plreq.fuzzitem = res.to_new_url(url)
        plreq.fuzzitem.payload = [FuzzPayload(url, [None])]

        return plreq



import unittest

# Python 2 and 3: urlib.parse

from wfuzz.fuzzobjects import FuzzRequest
from wfuzz.fuzzobjects import FuzzResult
from wfuzz.filter import FuzzResFilter


raw_req = """GET / HTTP/1.1
Host: www.wfuzz.org
User-Agent: curl/7.58.0
Accept: */*
"""

raw_resp = b"""HTTP/1.1 302 Found
Content-Type: text/html; charset=utf-8
Content-Language: en
Location: https://wfuzz.readthedocs.io/en/latest/
Vary: Accept-Language, Cookie
Server: nginx/1.14.0 (Ubuntu)
X-Fallback: True
X-Served: Django
X-Deity: web01
Date: Wed, 23 Jan 2019 21:43:59 GMT
Content-Length: 0
"""


class FilterTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(FilterTest, self).__init__(*args, **kwargs)
        self.maxDiff = 1000

    def get_filtered_fuzzrequest(self, filter_str):
        fr = FuzzRequest()
        fr.update_from_raw_http(raw_req, "http", raw_resp, b"")

        fuzz_res = FuzzResult(history=fr)

        ffilter = FuzzResFilter(filter_string=filter_str)
        ffilter.is_visible(fuzz_res)

        return fuzz_res

    def test_code_set(self):
        self.assertEqual(self.get_filtered_fuzzrequest("r.code:=429").code, 429)
        self.assertEqual(self.get_filtered_fuzzrequest("r.c:=404").code, 404)
        self.assertEqual(self.get_filtered_fuzzrequest("r.c=+404").code, 706)
        self.assertEqual(self.get_filtered_fuzzrequest("r.c=-404").code, 706)

    def test_url_set(self):
        fr = FuzzRequest()
        fr.url = "http://www.wfuzz.org/path?param=1&param2=2"

        fuzz_res = FuzzResult(history=fr)

        ffilter = FuzzResFilter(filter_string="r.url=+'test'")
        ffilter.is_visible(fuzz_res)
        self.assertEqual(fuzz_res.history.url, "http://www.wfuzz.org/path?param=1&param2=2test")

        ffilter = FuzzResFilter(filter_string="r.url:='test'")
        ffilter.is_visible(fuzz_res)
        self.assertEqual(fuzz_res.history.url, "http://test/")

        ffilter = FuzzResFilter(filter_string="r.url=-'test'")
        ffilter.is_visible(fuzz_res)
        self.assertEqual(fuzz_res.history.url, "testhttp://test/")

    def test_nonexisting(self):
        fr = FuzzRequest()
        fr.url = "http://www.wfuzz.org/path?param=1&param2=2"

        fuzz_res = FuzzResult(history=fr)

        with self.assertRaises(Exception) as context:
            ffilter = FuzzResFilter(filter_string="url=-'test'")
            ffilter.is_visible(fuzz_res)
            self.assertTrue("rsetattr: Can't set" in str(context.exception))

        with self.assertRaises(Exception) as context:
            ffilter = FuzzResFilter(filter_string="notthere=-'test'")
            ffilter.is_visible(fuzz_res)
            self.assertTrue("rgetattr: Can't get" in str(context.exception))

        with self.assertRaises(Exception) as context:
            ffilter = FuzzResFilter(filter_string="r.params.get.notthere=-'test'")
            ffilter.is_visible(fuzz_res)
            self.assertTrue("DotDict: Non-existing field" in str(context.exception))

    def test_params_set_no_value(self):
        fr = FuzzRequest()
        fr.url = "http://www.wfuzz.org/path?param"

        fuzz_res = FuzzResult(history=fr)

        ffilter = FuzzResFilter(filter_string="r.params.all=+'test'")
        ffilter.is_visible(fuzz_res)
        self.assertEqual(fuzz_res.history.params.get, {'param': None})

    def test_params_set(self):
        fr = FuzzRequest()
        fr.url = "http://www.wfuzz.org/path?param=1&param2=2"

        fuzz_res = FuzzResult(history=fr)

        ffilter = FuzzResFilter(filter_string="r.params.get.param=+'test'")
        ffilter.is_visible(fuzz_res)
        self.assertEqual(fuzz_res.history.params.get.param, "1test")
        self.assertEqual(fuzz_res.history.params.get, {'param': "1test", 'param2': "2"})

        ffilter = FuzzResFilter(filter_string="r.params.get.param=-'test'")
        ffilter.is_visible(fuzz_res)
        self.assertEqual(fuzz_res.history.params.get.param, "test1test")
        self.assertEqual(fuzz_res.history.params.get, {'param': "test1test", 'param2': "2"})

        ffilter = FuzzResFilter(filter_string="r.params.get.param:='test'")
        ffilter.is_visible(fuzz_res)
        self.assertEqual(fuzz_res.history.params.get.param, "test")
        self.assertEqual(fuzz_res.history.params.get, {'param': "test", 'param2': "2"})

        ffilter = FuzzResFilter(filter_string="r.params.get.param2='2'")
        self.assertEqual(ffilter.is_visible(fuzz_res), True)

        fr.url = "http://www.wfuzz.org/path?param=1&param2=2"
        ffilter = FuzzResFilter(filter_string="r.params.all=+'2'")
        ffilter.is_visible(fuzz_res)
        self.assertEqual(fuzz_res.history.params.all, {'param': "12", 'param2': "22"})

        fr.url = "http://www.wfuzz.org/path?param=1&param2=2"
        ffilter = FuzzResFilter(filter_string="r.params.all:='2'")
        ffilter.is_visible(fuzz_res)
        self.assertEqual(fuzz_res.history.params.all, {'param': "2", 'param2': "2"})

    def test_urlp(self):
        fr = FuzzRequest()
        fr.url = "http://www.wfuzz.org/path/test.php?param=1&param2=2"

        fuzz_res = FuzzResult(history=fr)

        ffilter = FuzzResFilter(filter_string="r.urlp.scheme='http'")
        self.assertEqual(True, ffilter.is_visible(fuzz_res))

        ffilter = FuzzResFilter(filter_string="r.urlp.netloc='www.wfuzz.org'")
        self.assertEqual(True, ffilter.is_visible(fuzz_res))

        ffilter = FuzzResFilter(filter_string="r.urlp.path='/path/test.php'")
        self.assertEqual(True, ffilter.is_visible(fuzz_res))

        ffilter = FuzzResFilter(filter_string="r.urlp.ffname='test.php'")
        self.assertEqual(True, ffilter.is_visible(fuzz_res))

        ffilter = FuzzResFilter(filter_string="r.urlp.fext='.php'")
        self.assertEqual(True, ffilter.is_visible(fuzz_res))

        ffilter = FuzzResFilter(filter_string="r.urlp.fname='test'")
        self.assertEqual(True, ffilter.is_visible(fuzz_res))

        ffilter = FuzzResFilter(filter_string="r.urlp.hasquery")
        self.assertEqual(True, ffilter.is_visible(fuzz_res))

        ffilter = FuzzResFilter(filter_string="not r.urlp.isbllist")
        self.assertEqual(True, ffilter.is_visible(fuzz_res))

    def test_ispath(self):
        fr = FuzzRequest()
        fr.url = "http://www.wfuzz.org/path?param=1&param2=2"
        fuzz_res = FuzzResult(history=fr)

        ffilter = FuzzResFilter(filter_string="r.is_path")
        self.assertEqual(False, ffilter.is_visible(fuzz_res))

        ffilter = FuzzResFilter(filter_string="r.pstrip")
        self.assertEqual(ffilter.is_visible(fuzz_res), "http://www.wfuzz.org/path-gparam-gparam2")

    def test_lwh(self):
        fr = FuzzRequest()
        fr.update_from_raw_http(raw_req, "http", raw_resp, b"Some line\n and words\nasdsdas")

        fuzz_res = FuzzResult(history=fr)

        ffilter = FuzzResFilter(filter_string="h=28 or w=6 or l=2")
        ffilter.is_visible(fuzz_res)
        self.assertEqual(True, ffilter.is_visible(fuzz_res))

    def test_location(self):
        fr = FuzzRequest()
        fr.update_from_raw_http(raw_req, "http", raw_resp, b"Some line\n and words\nasdsdas")

        fuzz_res = FuzzResult(history=fr)

        ffilter = FuzzResFilter(filter_string="r.headers.response.Location")
        ffilter.is_visible(fuzz_res)
        self.assertEqual('https://wfuzz.readthedocs.io/en/latest/', ffilter.is_visible(fuzz_res))



# slightly modified from
# https://gist.github.com/trungly/5889154

from http.server import HTTPServer
from http.server import SimpleHTTPRequestHandler
import urllib.parse


class GetHandler(SimpleHTTPRequestHandler):
    def do_HEAD(self):
        parsed_path = urllib.parse.urlparse(self.path)
        if parsed_path.path.startswith("/echo"):
            message = '\n'.join(
                [
                    'CLIENT VALUES:',
                    'client_address=%s (%s)' % (self.client_address, self.address_string()),
                    'command=%s' % self.command,
                    'path=%s' % self.path,
                    'real path=%s' % parsed_path.path,
                    'query=%s' % parsed_path.query,
                    'request_version=%s' % self.request_version,
                    '',
                    'HEADERS:',
                    '%s' % self.headers,
                ]
            )
            self.send_response(200)
            self.end_headers()
            self.wfile.write(message.encode('utf-8'))
        elif parsed_path.path.startswith("/redirect"):
            self.send_response(301)
            self.send_header('Location', "/echo")
            self.end_headers()
        else:
            SimpleHTTPRequestHandler.do_HEAD(self)

        return

    def do_GET(self):
        parsed_path = urllib.parse.urlparse(self.path)
        if parsed_path.path.startswith("/echo"):
            message = '\n'.join(
                [
                    'CLIENT VALUES:',
                    'client_address=%s (%s)' % (self.client_address, self.address_string()),
                    'command=%s' % self.command,
                    'path=%s' % self.path,
                    'real path=%s' % parsed_path.path,
                    'query=%s' % parsed_path.query,
                    'request_version=%s' % self.request_version,
                    '',
                    'HEADERS:',
                    '%s' % self.headers,
                ]
            )
            self.send_response(200)
            self.end_headers()
            self.wfile.write(message.encode('utf-8'))
        elif parsed_path.path.startswith("/redirect"):
            self.send_response(301)
            self.send_header('Location', "/echo")
            self.end_headers()
        else:
            SimpleHTTPRequestHandler.do_GET(self)

        return

    def do_POST(self):
        parsed_path = urllib.parse.urlparse(self.path)
        if parsed_path.path.startswith("/echo"):
            content_len = int(self.headers.get('content-length'))
            post_body = self.rfile.read(content_len).decode('utf-8')
            self.send_response(200)
            self.end_headers()

            message = '\n'.join(
                [
                    'CLIENT VALUES:',
                    'client_address=%s (%s)' % (self.client_address, self.address_string()),
                    'command=%s' % self.command,
                    'path=%s' % self.path,
                    'real path=%s' % parsed_path.path,
                    'query=%s' % parsed_path.query,
                    'request_version=%s' % self.request_version,
                    '',
                    'HEADERS:',
                    '%s' % self.headers,
                    'POST_DATA=%s' % post_body,
                    '',
                ]
            )

            self.wfile.write(message.encode('utf-8'))

        return


if __name__ == '__main__':
    server = HTTPServer(('0.0.0.0', 8000), GetHandler)
    server.serve_forever()



#!/usr/bin/env python
import sys

from .core import Fuzzer
from .facade import Facade
from .exception import FuzzException, FuzzExceptBadInstall

from .ui.console.mvc import Controller, KeyPress, View
from .ui.console.common import help_banner2
from .ui.console.clparser import CLParser

from .fuzzobjects import FuzzResult


def main():
    kb = None
    fz = None
    printer = None
    session_options = None

    try:
        # parse command line
        session_options = CLParser(sys.argv).parse_cl().compile()
        session_options["send_discarded"] = True

        # Create fuzzer's engine
        fz = Fuzzer(session_options)

        if session_options["interactive"]:
            # initialise controller
            try:
                kb = KeyPress()
            except ImportError as e:
                raise FuzzExceptBadInstall("Error importing necessary modules for interactive mode: %s" % str(e))
            else:
                Controller(fz, kb)
                kb.start()

        printer = View(session_options)
        if session_options["console_printer"]:
            printer = Facade().printers.get_plugin(session_options["console_printer"])(None)
        printer.header(fz.genReq.stats)

        for res in fz:
            printer.result(res)

        printer.footer(fz.genReq.stats)
    except FuzzException as e:
        print("\nFatal exception: {}".format(str(e)))
    except KeyboardInterrupt:
        print("\nFinishing pending requests...")
        if fz:
            fz.cancel_job()
    except NotImplementedError as e:
        print("\nFatal exception: Error importing wfuzz extensions: {}".format(str(e)))
    except Exception as e:
        print("\nUnhandled exception: {}".format(str(e)))
    finally:
        if session_options:
            session_options.close()
        if kb:
            kb.cancel_job()
        Facade().sett.save()


def main_filter():
    def usage():
        print(help_banner2)
        print("""Usage:
\n\twfpayload [Options]\n\n
\nOptions:\n
\t--help              : This help
\t-v                  : Verbose output
\t-z payload          : Specify a payload for each FUZZ keyword used in the form of type,parameters,encoder.
\t		      A list of encoders can be used, ie. md5-sha1. Encoders can be chained, ie. md5@sha1.
\t		      Encoders category can be used. ie. url
\t--zD default	    : Default argument for the specified payload (it must be preceded by -z or -w).
\t--zP <params>	    : Arguments for the specified payload (it must be preceded by -z or -w).
\t--slice <filter>    : Filter payload\'s elements using the specified expression.
\t-w wordlist         : Specify a wordlist file (alias for -z file,wordlist).
\t-m iterator         : Specify an iterator for combining payloads (product by default)
\t--field <expr>      : Do not show the payload but the specified language expression
\t--efield <expr>     : Show the specified language expression together with the current payload
""")

    from .api import payload
    from .exception import FuzzExceptBadOptions
    import getopt

    try:
        opts, args = getopt.getopt(sys.argv[1:], "vhz:m:w:", ["field=", "help", "slice=", "zD=", "zP=", "efield="])
    except getopt.GetoptError as err:
        print((str(err)))
        usage()
        sys.exit(2)

    if len(opts) == 0 or len(args) > 0:
        usage()
        sys.exit()

    field = None
    raw_output = False

    for o, value in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        if o in ("--efield"):
            field = value
        if o in ("--field"):
            field = value
            raw_output = True

    try:
        session_options = CLParser(sys.argv).parse_cl()
        printer = None

        for res in payload(**session_options):
            if len(res) > 1:
                raise FuzzExceptBadOptions("wfpayload can only be used to generate one word dictionaries")
            else:
                r = res[0]

            # TODO: all should be same object type and no need for isinstance
            if isinstance(r, FuzzResult):
                if raw_output:
                    print(r.eval(field if field is not None else "url"))
                else:
                    if printer is None:
                        printer = View(session_options)
                        printer.header(None)

                    if field:
                        r._description = field
                        r._show_field = False
                    printer.result(r)
            else:
                print(r)

    except KeyboardInterrupt:
        pass
    except FuzzException as e:
        print(("\nFatal exception: %s" % str(e)))
    except Exception as e:
        print(("\nUnhandled exception: %s" % str(e)))


def main_encoder():
    def usage():
        print(help_banner2)
        print("Usage:")
        print("\n\twfencode --help This help")
        print("\twfencode -d decoder_name string_to_decode")
        print("\twfencode -e encoder_name string_to_encode")
        print()

    from .api import encode, decode
    import getopt

    try:
        opts, args = getopt.getopt(sys.argv[1:], "he:d:", ["help"])
    except getopt.GetoptError as err:
        print((str(err)))
        usage()
        sys.exit(2)

    if len(args) == 0:
        usage()
        sys.exit()

    try:
        for o, value in opts:
            if o == "-e":
                print((encode(value, args[0])))
            elif o == "-d":
                print((decode(value, args[0])))
            elif o in ("-h", "--help"):
                usage()
                sys.exit()
    except IndexError as e:
        usage()
        print("\nFatal exception: Specify a string to encode or decode.{}\n".format(str(e)))
        sys.exit()
    except AttributeError as e:
        print("\nEncoder plugin missing encode or decode functionality. {}".format(str(e)))
    except FuzzException as e:
        print(("\nFatal exception: %s" % str(e)))


def main_gui():
    import wx
    from .ui.gui.guicontrols import WfuzzFrame
    from .ui.gui.controller import GUIController

    app = wx.App(False)

    frame = WfuzzFrame(None, wx.ID_ANY, "WFuzz wxPython Console", size=(750, 590))
    GUIController(frame)

    frame.Show()
    app.MainLoop()



import pycurl
from io import BytesIO
from threading import Thread, Lock
import itertools
from queue import Queue

from .exception import FuzzExceptBadOptions, FuzzExceptNetError


class HttpPool:
    HTTPAUTH_BASIC, HTTPAUTH_NTLM, HTTPAUTH_DIGEST = ('basic', 'ntlm', 'digest')
    newid = itertools.count(0)

    def __init__(self, options):
        self.processed = 0

        self.exit_job = False
        self.mutex_multi = Lock()
        self.mutex_stats = Lock()
        self.mutex_reg = Lock()

        self.m = None
        self.freelist = Queue()
        self.retrylist = Queue()

        self.ths = None

        self.pool_map = {}
        self.default_poolid = 0

        self.options = options

        self._registered = 0

    def _initialize(self):
        # pycurl Connection pool
        self._create_pool(self.options.get("concurrent"))

        # internal pool
        self.default_poolid = self._new_pool()

        # create threads
        self.ths = []

        for fn in ("_read_multi_stack", "_read_retry_queue"):
            th = Thread(target=getattr(self, fn))
            th.setName(fn)
            self.ths.append(th)
            th.start()

    def job_stats(self):
        with self.mutex_stats:
            dic = {
                "http_Processed": self.processed,
                "http_Idle Workers": self.freelist.qsize()
            }
        return dic

    # internal http pool control

    def perform(self, fuzzreq):
        poolid = self._new_pool()
        self.enqueue(fuzzreq, poolid)
        item = self.pool_map[poolid]["queue"].get()
        return item

    def iter_results(self, poolid=None):
        item = self.pool_map[self.default_poolid if not poolid else poolid]["queue"].get()

        if not item:
            return

        yield item

    def _new_pool(self):
        poolid = next(self.newid)
        self.pool_map[poolid] = {}
        self.pool_map[poolid]["queue"] = Queue()
        self.pool_map[poolid]["proxy"] = None

        if self.options.get("proxies"):
            self.pool_map[poolid]["proxy"] = self._get_next_proxy(self.options.get("proxies"))

        return poolid

    def enqueue(self, fuzzres, poolid=None):
        c = fuzzres.history.to_http_object(self.freelist.get())
        c = self._set_extra_options(c, fuzzres, self.default_poolid if not poolid else poolid)

        if self.exit_job:
            return

        c.response_queue = ((BytesIO(), BytesIO(), fuzzres, self.default_poolid if not poolid else poolid))
        c.setopt(pycurl.WRITEFUNCTION, c.response_queue[0].write)
        c.setopt(pycurl.HEADERFUNCTION, c.response_queue[1].write)

        with self.mutex_multi:
            self.m.add_handle(c)

    def _stop_to_pools(self):
        for p in list(self.pool_map.keys()):
            self.pool_map[p]["queue"].put(None)

    # Pycurl management
    def _create_pool(self, num_conn):
        # Pre-allocate a list of curl objects
        self.m = pycurl.CurlMulti()
        self.m.handles = []

        for i in range(num_conn):
            c = pycurl.Curl()
            self.m.handles.append(c)
            self.freelist.put(c)

    def cleanup(self):
        self.exit_job = True
        for th in self.ths:
            th.join()

    def register(self):
        with self.mutex_reg:
            self._registered += 1

            if not self.pool_map:
                self._initialize()
                return self.default_poolid
            else:
                return self._new_pool()

    def deregister(self):
        with self.mutex_reg:
            self._registered -= 1

        if self._registered <= 0:
            self.cleanup()

    def _get_next_proxy(self, proxy_list):
        i = 0
        while 1:
            yield proxy_list[i]
            i += 1
            i = i % len(proxy_list)

    def _set_extra_options(self, c, freq, poolid):
        if self.pool_map[poolid]["proxy"]:
            ip, port, ptype = next(self.pool_map[poolid]["proxy"])

            freq.wf_proxy = (("%s:%s" % (ip, port)), ptype)

            c.setopt(pycurl.PROXY, "%s:%s" % (ip, port))
            if ptype == "SOCKS5":
                c.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_SOCKS5)
            elif ptype == "SOCKS4":
                c.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_SOCKS4)
            elif ptype == "HTTP":
                pass
            else:
                raise FuzzExceptBadOptions("Bad proxy type specified, correct values are HTTP, SOCKS4 or SOCKS5.")
        else:
            c.setopt(pycurl.PROXY, "")

        mdelay = self.options.get("req_delay")
        if mdelay is not None:
            c.setopt(pycurl.TIMEOUT, mdelay)

        cdelay = self.options.get("conn_delay")
        if cdelay is not None:
            c.setopt(pycurl.CONNECTTIMEOUT, cdelay)

        return c

    def _read_retry_queue(self):
        while not self.exit_job:
            res, poolid = self.retrylist.get()

            if res is None:
                break

            self.enqueue(res, poolid)

    def _read_multi_stack(self):
        # Check for curl objects which have terminated, and add them to the freelist
        while not self.exit_job:
            with self.mutex_multi:
                while not self.exit_job:
                    ret, num_handles = self.m.perform()
                    if ret != pycurl.E_CALL_MULTI_PERFORM:
                        break

            num_q, ok_list, err_list = self.m.info_read()
            for c in ok_list:
                # Parse response
                buff_body, buff_header, res, poolid = c.response_queue

                try:
                    res.history.from_http_object(c, buff_header.getvalue(), buff_body.getvalue())
                except Exception as e:
                    self.pool_map[poolid]["queue"].put(res.update(exception=e))
                else:
                    # reset type to result otherwise backfeed items will enter an infinite loop
                    self.pool_map[poolid]["queue"].put(res.update())

                self.m.remove_handle(c)
                self.freelist.put(c)

                with self.mutex_stats:
                    self.processed += 1

            for c, errno, errmsg in err_list:
                buff_body, buff_header, res, poolid = c.response_queue

                res.history.totaltime = 0
                self.m.remove_handle(c)
                self.freelist.put(c)

                # Usual suspects:

                # Exception in perform (35, 'error:0B07C065:x509 certificate routines:X509_STORE_add_cert:cert already in hash table')
                # Exception in perform (18, 'SSL read: error:0B07C065:x509 certificate routines:X509_STORE_add_cert:cert already in hash table, errno 11')
                # Exception in perform (28, 'Connection time-out')
                # Exception in perform (7, "couldn't connect to host")
                # Exception in perform (6, "Couldn't resolve host 'www.xxx.com'")
                # (28, 'Operation timed out after 20000 milliseconds with 0 bytes received')
                # Exception in perform (28, 'SSL connection timeout')
                # 5 Couldn't resolve proxy 'aaa'

                # retry requests with recoverable errors
                if errno not in [28, 7, 6, 5]:
                    res.history.wf_retries += 1

                    if res.history.wf_retries < self.options.get("retries"):
                        self.retrylist.put((res, poolid))
                        continue

                e = FuzzExceptNetError("Pycurl error %d: %s" % (errno, errmsg))
                self.pool_map[poolid]["queue"].put(res.update(exception=e))

                with self.mutex_stats:
                    self.processed += 1

        self._stop_to_pools()
        self.retrylist.put((None, None))
        # cleanup multi stack
        for c in self.m.handles:
            c.close()
            self.freelist.put(c)
        self.m.close()



import unittest

from wfuzz.ui.console.clparser import CLParser


class CLParserTest(unittest.TestCase):
    def test_listplugins(self):
        with self.assertRaises(SystemExit) as cm:
            CLParser(['wfuzz', '-e', 'iterators']).parse_cl()

        self.assertEqual(cm.exception.code, 0)

    def test_ip_option(self):
        options = CLParser(['wfuzz', '--ip', '127.0.0.1']).parse_cl()

        self.assertEqual(options.data['connect_to_ip']['ip'], '127.0.0.1')
        self.assertEqual(options.data['connect_to_ip']['port'], '80')

        options = CLParser(['wfuzz', '--ip', '127.0.0.1:22']).parse_cl()

        self.assertEqual(options.data['connect_to_ip']['ip'], '127.0.0.1')
        self.assertEqual(options.data['connect_to_ip']['port'], '22')

        options = CLParser(['wfuzz', '--ip', '127.0.0.1:']).parse_cl()

        self.assertEqual(options.data['connect_to_ip']['ip'], '127.0.0.1')
        self.assertEqual(options.data['connect_to_ip']['port'], '80')

        with self.assertRaises(Exception) as cm:
            options = CLParser(['wfuzz', '--ip', ':80']).parse_cl()
        self.assertTrue("An IP must be specified" in str(cm.exception))

    def test_ze_zd_option(self):
        with self.assertRaises(Exception) as cm:
            options = CLParser(['wfuzz', '-z', 'range,0-10', '--zD', '0-10', 'url']).parse_cl()
        self.assertTrue("exclusive" in str(cm.exception))

        options = CLParser(['wfuzz', '-z', 'range', '--zD', '0-1', '--zE', 'md5', 'url']).parse_cl()
        self.assertEqual(options.data['payloads'], [('range', {'default': '0-1', 'encoder': ['md5']}, None)])

        options = CLParser(['wfuzz', '-z', 'range,0-1', '--zE', 'md5', 'url']).parse_cl()
        self.assertEqual(options.data['payloads'], [('range', {'default': '0-1', 'encoder': ['md5']}, None)])

        options = CLParser(['wfuzz', '-z', 'range', '--zD', '0-1', '--zE', 'md5', 'url']).parse_cl()
        self.assertEqual(options.data['payloads'], [('range', {'default': '0-1', 'encoder': ['md5']}, None)])

        options = CLParser(['wfuzz', '-z', 'range', '--zD', '0-1']).parse_cl()
        self.assertEqual(options.data['payloads'], [('range', {'default': '0-1', 'encoder': None}, None)])

        options = CLParser(['wfuzz', '-z', 'range,0-1']).parse_cl()
        self.assertEqual(options.data['payloads'], [('range', {'default': '0-1', 'encoder': None}, None)])



from . import utils
from . import __version__ as version
from .externals.moduleman.registrant import MulRegistrant
from .externals.moduleman.loader import DirLoader
from .externals.settings.settings import SettingsBase
from .exception import FuzzExceptNoPluginError, FuzzExceptPluginLoadError

import os

# python2 and 3: metaclass
from future.utils import with_metaclass

ERROR_CODE = -1
BASELINE_CODE = -2


class Settings(SettingsBase):
    def get_config_file(self):
        return os.path.join(utils.get_home(check=True), "wfuzz.ini")

    def set_defaults(self):
        return dict(
            plugins=[
                ("bing_apikey", ''),
                ("shodan_apikey", '')
            ],
            kbase=[("discovery.blacklist", '.svg-.css-.js-.jpg-.gif-.png-.jpeg-.mov-.avi-.flv-.ico')],
            connection=[
                ("concurrent", '10'),
                ("conn_delay", '90'),
                ("req_delay", '90'),
                ("retries", '3'),
                ("User-Agent", "Wfuzz/%s" % version)
            ],
            general=[
                ("default_printer", 'raw'),
                ("cancel_on_plugin_except", "0"),
                ("concurrent_plugins", '3'),
                ("lookup_dirs", '.'),
                ("encode_space", '1')
            ],
        )


class MyRegistrant(MulRegistrant):
    def get_plugin(self, identifier):
        try:
            return MulRegistrant.get_plugin(self, identifier)
        except KeyError as e:
            raise FuzzExceptNoPluginError("Requested plugin %s. Error: %s" % (identifier, str(e)))


# python2 and 3: class Facade(metaclass=utils.Singleton):
class Facade(with_metaclass(utils.Singleton, object)):
    def __init__(self):

        self.__plugins = dict(
            printers=None,
            scripts=None,
            encoders=None,
            iterators=None,
            payloads=None,
        )

        self.sett = Settings()

    def _load(self, cat):
        try:
            if cat not in self.__plugins:
                raise FuzzExceptNoPluginError("Non-existent plugin category %s" % cat)

            if not self.__plugins[cat]:
                loader_list = []
                loader_list.append(DirLoader(**{"base_dir": cat, "base_path": utils.get_path("plugins")}))
                loader_list.append(DirLoader(**{"base_dir": cat, "base_path": utils.get_home()}))
                self.__plugins[cat] = MyRegistrant(loader_list)

            return self.__plugins[cat]
        except Exception as e:
            raise FuzzExceptPluginLoadError("Error loading plugins: %s" % str(e))

    def proxy(self, which):
        return self._load(which)

    def get_registrants(self):
        return self.__plugins.keys()

    def __getattr__(self, name):
        if name in ["printers", "payloads", "iterators", "encoders", "scripts"]:
            return self._load(name)
        else:
            raise AttributeError



from .exception import FuzzExceptIncorrectFilter, FuzzExceptBadOptions, FuzzExceptInternalError, FuzzException
from .utils import rgetattr, rsetattr, value_in_any_list_item

import re
import collections
import operator

# Python 2 and 3: alternative 4
try:
    from urllib.parse import unquote
except ImportError:
    from urllib import unquote

from .facade import Facade, ERROR_CODE, BASELINE_CODE


PYPARSING = True
try:
    from pyparsing import Word, Group, oneOf, Optional, Suppress, ZeroOrMore, Literal, alphas, QuotedString
    from pyparsing import ParseException
except ImportError:
    PYPARSING = False


class FuzzResFilter:
    def __init__(self, ffilter=None, filter_string=None):
        if PYPARSING:
            quoted_str_value = QuotedString('\'', unquoteResults=True, escChar='\\')
            int_values = Word("0123456789").setParseAction(lambda s, l, t: [int(t[0])])
            error_value = Literal("XXX").setParseAction(self.__compute_xxx_value)
            bbb_value = Literal("BBB").setParseAction(self.__compute_bbb_value)
            field_value = Word(alphas + "." + "_" + "-")
            reserverd_words = oneOf("BBB XXX")

            basic_primitives = int_values | quoted_str_value

            operator_names = oneOf("m d e un u r l sw gre gregex unique startswith decode encode unquote replace lower upper").setParseAction(lambda s, l, t: [(l, t[0])])

            fuzz_symbol = (Suppress("FUZ") + Optional(Word("23456789"), 1).setParseAction(lambda s, l, t: [int(t[0]) - 1]) + Suppress("Z")).setParseAction(self._compute_fuzz_symbol)
            operator_call = Group(Suppress("|") + operator_names + Suppress("(") + Optional(basic_primitives, None) + Optional(Suppress(",") + basic_primitives, None) + Suppress(")"))

            fuzz_value = (fuzz_symbol + Optional(Suppress("[") + field_value + Suppress("]"), None)).setParseAction(self.__compute_fuzz_value)
            fuzz_value_op = ((fuzz_symbol + Suppress("[") + Optional(field_value)).setParseAction(self.__compute_fuzz_value) + operator_call + Suppress("]")).setParseAction(self.__compute_perl_value)
            fuzz_value_op2 = ((fuzz_symbol + operator_call).setParseAction(self.__compute_perl_value))

            res_value_op = (~reserverd_words + Word("0123456789" + alphas + "." + "_" + "-").setParseAction(self.__compute_res_value) + Optional(operator_call, None)).setParseAction(self.__compute_perl_value)
            basic_primitives_op = (basic_primitives + Optional(operator_call, None)).setParseAction(self.__compute_perl_value)

            fuzz_statement = basic_primitives_op ^ fuzz_value ^ fuzz_value_op ^ fuzz_value_op2 ^ res_value_op

            operator = oneOf("and or")
            not_operator = Optional(oneOf("not"), "notpresent")

            symbol_expr = Group(fuzz_statement + oneOf("= == != < > >= <= =~ !~ ~ := =+ =-") + (bbb_value ^ error_value ^ basic_primitives ^ fuzz_statement)).setParseAction(self.__compute_expr)

            definition = symbol_expr ^ fuzz_statement
            definition_not = not_operator + definition
            definition_expr = definition_not + ZeroOrMore(operator + definition_not)

            nested_definition = Group(Suppress("(") + definition_expr + Suppress(")"))
            nested_definition_not = not_operator + nested_definition

            self.finalformula = (nested_definition_not ^ definition_expr) + ZeroOrMore(operator + (nested_definition_not ^ definition_expr))

            definition_not.setParseAction(self.__compute_not_operator)
            nested_definition_not.setParseAction(self.__compute_not_operator)
            nested_definition.setParseAction(self.__compute_formula)
            self.finalformula.setParseAction(self.__myreduce)

        if ffilter is not None and filter_string is not None:
            raise FuzzExceptInternalError(FuzzException.FATAL, "A filter must be initilized with a filter string or an object, not both")

        self.res = None
        if ffilter:
            self.hideparams = ffilter
        else:
            self.hideparams = dict(
                regex_show=None,
                codes_show=None,
                codes=[],
                words=[],
                lines=[],
                chars=[],
                regex=None,
                filter_string=""
            )

        if filter_string:
            self.hideparams['filter_string'] = filter_string

        self.baseline = None
        self.stack = []

        self._cache = collections.defaultdict(set)

    def set_baseline(self, res):
        if BASELINE_CODE in self.hideparams['lines']:
            self.hideparams['lines'].append(res.lines)
        if BASELINE_CODE in self.hideparams['codes']:
            self.hideparams['codes'].append(res.code)
        if BASELINE_CODE in self.hideparams['words']:
            self.hideparams['words'].append(res.words)
        if BASELINE_CODE in self.hideparams['chars']:
            self.hideparams['chars'].append(res.chars)

        self.baseline = res

    def __compute_res_value(self, tokens):
        self.stack.append(tokens[0])

        try:
            return rgetattr(self.res, tokens[0])
        except AttributeError:
            raise FuzzExceptIncorrectFilter("Non-existing introspection field or HTTP parameter \"{}\"!".format(tokens[0]))

    def _compute_fuzz_symbol(self, tokens):
        i = tokens[0]

        try:
            return self.res.payload[i].content
        except IndexError:
            raise FuzzExceptIncorrectFilter("Non existent FUZZ payload! Use a correct index.")

    def __compute_fuzz_value(self, tokens):
        fuzz_val, field = tokens

        self.stack.append(field)

        try:
            return rgetattr(fuzz_val, field) if field else fuzz_val
        except IndexError:
            raise FuzzExceptIncorrectFilter("Non existent FUZZ payload! Use a correct index.")
        except AttributeError as e:
            raise FuzzExceptIncorrectFilter("A field expression must be used with a fuzzresult payload not a string. %s" % str(e))

    def __compute_bbb_value(self, tokens):
        element = self.stack[0] if self.stack else None

        if self.baseline is None:
            raise FuzzExceptBadOptions("FilterQ: specify a baseline value when using BBB")

        if element == 'l' or element == 'lines':
            ret = self.baseline.lines
        elif element == 'c' or element == 'code':
            ret = self.baseline.code
        elif element == 'w' or element == 'words':
            ret = self.baseline.words
        elif element == 'h' or element == 'chars':
            return self.baseline.chars
        elif element == 'index' or element == 'i':
            ret = self.baseline.nres

        return str(ret)

    def __compute_perl_value(self, tokens):
        leftvalue, exp = tokens
        # import pdb; pdb.set_trace()

        if exp:
            loc_op, middlevalue, rightvalue = exp
            loc, op = loc_op
        else:
            return leftvalue

        if (op == "un" or op == "unquote") and middlevalue is None and rightvalue is None:
            ret = unquote(leftvalue)
        elif (op == "e" or op == "encode") and middlevalue is not None and rightvalue is None:
            ret = Facade().encoders.get_plugin(middlevalue)().encode(leftvalue)
        elif (op == "d" or op == "decode") and middlevalue is not None and rightvalue is None:
            ret = Facade().encoders.get_plugin(middlevalue)().decode(leftvalue)
        elif op == "r" or op == "replace":
            return leftvalue.replace(middlevalue, rightvalue)
        elif op == "upper":
            return leftvalue.upper()
        elif op == "lower" or op == "l":
            return leftvalue.lower()
        elif op == 'gregex' or op == "gre":
            search_res = None
            try:
                regex = re.compile(middlevalue)
                search_res = regex.search(leftvalue)
            except re.error as e:
                raise FuzzExceptBadOptions("Invalid regex expression used in expression: %s" % str(e))

            if search_res is None:
                return ''
            return search_res.group(1)
        elif op == 'startswith' or op == "sw":
            return leftvalue.strip().startswith(middlevalue)
        elif op == 'unique' or op == "u":
            if leftvalue not in self._cache[loc]:
                self._cache[loc].add(leftvalue)
                return True
            else:
                return False
        else:
            raise FuzzExceptBadOptions("Bad format, expression should be m,d,e,r,s(value,value)")

        return ret

    def __compute_xxx_value(self, tokens):
        return ERROR_CODE

    def __compute_expr(self, tokens):
        leftvalue, exp_operator, rightvalue = tokens[0]

        field_to_set = self.stack[0] if self.stack else None

        try:
            if exp_operator in ["=", '==']:
                return str(leftvalue) == str(rightvalue)
            elif exp_operator == "<=":
                return leftvalue <= rightvalue
            elif exp_operator == ">=":
                return leftvalue >= rightvalue
            elif exp_operator == "<":
                return leftvalue < rightvalue
            elif exp_operator == ">":
                return leftvalue > rightvalue
            elif exp_operator == "!=":
                return leftvalue != rightvalue
            elif exp_operator == "=~":
                regex = re.compile(rightvalue, re.MULTILINE | re.DOTALL)
                return regex.search(leftvalue) is not None
            elif exp_operator in ["!~", "~"]:
                ret = True

                if isinstance(leftvalue, str):
                    ret = rightvalue.lower() in leftvalue.lower()
                elif isinstance(leftvalue, list):
                    ret = value_in_any_list_item(rightvalue, leftvalue)
                elif isinstance(leftvalue, dict):
                    return len({k: v for (k, v) in leftvalue.items() if rightvalue.lower() in k.lower() or value_in_any_list_item(rightvalue, v)}) > 0
                else:
                    raise FuzzExceptBadOptions("Invalid operand type {}".format(rightvalue))

                return ret if exp_operator == "~" else not ret
            elif exp_operator == ":=":
                rsetattr(self.res, field_to_set, rightvalue, None)
            elif exp_operator == "=+":
                rsetattr(self.res, field_to_set, rightvalue, operator.add)
            elif exp_operator == "=-":
                rsetattr(self.res, field_to_set, rightvalue, lambda x, y: y + x)
        except re.error as e:
            raise FuzzExceptBadOptions("Invalid regex expression used in expression: %s" % str(e))
        except TypeError as e:
            raise FuzzExceptBadOptions("Invalid operand types used in expression: %s" % str(e))
        except ParseException as e:
            raise FuzzExceptBadOptions("Invalid filter: %s" % str(e))

        return True

    def __myreduce(self, elements):
        first = elements[0]
        for i in range(1, len(elements), 2):
            if elements[i] == "and":
                first = (first and elements[i + 1])
            elif elements[i] == "or":
                first = (first or elements[i + 1])

        self.stack = []
        return first

    def __compute_not_operator(self, tokens):
        operator, value = tokens

        if operator == "not":
            return not value

        return value

    def __compute_formula(self, tokens):
        return self.__myreduce(tokens[0])

    def is_active(self):
        return any([
            self.hideparams['regex_show'] is not None,
            self.hideparams['codes_show'] is not None,
            self.hideparams['filter_string'] != "",
        ])

    def is_visible(self, res):
        filter_string = self.hideparams['filter_string']
        if filter_string and PYPARSING:
            self.res = res
            try:
                return self.finalformula.parseString(filter_string, parseAll=True)[0]
            except ParseException as e:
                raise FuzzExceptIncorrectFilter("Incorrect filter expression, check documentation. {}".format(str(e)))
            except AttributeError as e:
                raise FuzzExceptIncorrectFilter("It is only possible to use advanced filters when using a non-string payload. %s" % str(e))
        else:
            if self.hideparams['codes_show'] is None:
                cond1 = True
            else:
                cond1 = not self.hideparams['codes_show']

            if self.hideparams['regex_show'] is None:
                cond2 = True
            else:
                cond2 = not self.hideparams['regex_show']

            if res.code in self.hideparams['codes'] or res.lines in self.hideparams['lines'] \
               or res.words in self.hideparams['words'] or res.chars in self.hideparams['chars']:
                cond1 = self.hideparams['codes_show']

            if self.hideparams['regex']:
                if self.hideparams['regex'].search(res.history.content):
                    cond2 = self.hideparams['regex_show']

            return (cond1 and cond2)

    @staticmethod
    def from_options(filter_options):
        ffilter = FuzzResFilter()

        ffilter.hideparams["filter_string"] = filter_options["filter"]

        try:
            if filter_options["ss"] is not None:
                ffilter.hideparams['regex_show'] = True
                ffilter.hideparams['regex'] = re.compile(filter_options['ss'], re.MULTILINE | re.DOTALL)

            elif filter_options["hs"] is not None:
                ffilter.hideparams['regex_show'] = False
                ffilter.hideparams['regex'] = re.compile(filter_options['hs'], re.MULTILINE | re.DOTALL)
        except Exception as e:
            raise FuzzExceptBadOptions("Invalid regex expression used in filter: %s" % str(e))

        if [x for x in ["sc", "sw", "sh", "sl"] if len(filter_options[x]) > 0]:
            ffilter.hideparams['codes_show'] = True
            ffilter.hideparams['codes'] = filter_options["sc"]
            ffilter.hideparams['words'] = filter_options["sw"]
            ffilter.hideparams['lines'] = filter_options["sl"]
            ffilter.hideparams['chars'] = filter_options["sh"]
        elif [x for x in ["hc", "hw", "hh", "hl"] if len(filter_options[x]) > 0]:
            ffilter.hideparams['codes_show'] = False
            ffilter.hideparams['codes'] = filter_options["hc"]
            ffilter.hideparams['words'] = filter_options["hw"]
            ffilter.hideparams['lines'] = filter_options["hl"]
            ffilter.hideparams['chars'] = filter_options["hh"]

        return ffilter

    def get_fuzz_words(self):
        marker_regex = re.compile(r"FUZ\d*Z", re.MULTILINE | re.DOTALL)
        fuzz_words = marker_regex.findall(self.hideparams["filter_string"])

        return fuzz_words


class FuzzResFilterSlice(FuzzResFilter):
    def _compute_fuzz_symbol(self, tokens):
        i = tokens[0]

        if i != 0:
            raise FuzzExceptIncorrectFilter("Non existent FUZZ payload! Use a correct index.")

        return self.res



import unittest

from wfuzz.utils import DotDict


class FilterDotDict(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(FilterDotDict, self).__init__(*args, **kwargs)
        self.maxDiff = 1000

    def test_code_set(self):
        dd = DotDict({'a': '1'})
        dd2 = DotDict({'a': '2'})

        self.assertEqual(dd + "test", {'a': "1test"})
        self.assertEqual("test" + dd, {'a': "test1"})
        self.assertEqual(dd + dd2, {'a': "2"})



import sys
import re
from setuptools import setup, find_packages

with open("README.md", "rb") as f:
    long_descr = f.read().decode("utf-8")


version = re.search(
    r'^__version__\s*=\s*"(.*)"',
    open('src/wfuzz/__init__.py').read(),
    re.M
).group(1)


dev_requires = [
    'mock',
    'netaddr',
    'pip-tools',
]

install_requires = [
    'pycurl<=7.43.0.3',
    'pyparsing',
    'future',
    'six',
    'configparser;python_version<"3.5"',
    'chardet',
]

if sys.platform.startswith("win"):
    install_requires += ["colorama"]

setup(
    name="wfuzz",
    include_package_data=True,
    data_files=[('docs/user', ['docs/user/advanced.rst'])],
    packages=find_packages(where='src'),
    package_dir={'wfuzz': 'src/wfuzz'},
    entry_points={
        'console_scripts': [
            'wfuzz = wfuzz.wfuzz:main',
            'wfpayload = wfuzz.wfuzz:main_filter',
            'wfencode = wfuzz.wfuzz:main_encoder',
        ],
        'gui_scripts': [
            'wxfuzz = wfuzz.wfuzz:main_gui',
        ]
    },
    version=version,
    description="Wfuzz - The web fuzzer",
    long_description=long_descr,
    author="Xavi Mendez (@x4vi_mendez)",
    author_email="xmendez@edge-security.com",
    url="http://wfuzz.org",
    install_requires=install_requires,
    extras_require={
        'dev': dev_requires,
    },
    python_requires=">=2.6",
    classifiers=(
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ),
)



from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.exception import FuzzExceptBadFile, FuzzExceptBadOptions
from wfuzz.fuzzobjects import FuzzResult, FuzzRequest
from wfuzz.plugin_api.base import BasePayload
from wfuzz.utils import rgetattr


import datetime
import string
import re
import struct
import zipfile

TAG = re.compile(r'</?(\w*)>', re.M)  # Match a XML tag
nvprint = string.printable.replace('\x0b', '').replace('\x0c', '')  # Printables


@moduleman_plugin
class burpstate(BasePayload):
    name = "burpstate"
    author = ("Xavi Mendez (@xmendez)", )
    version = "0.1"
    description = (
        "*ALERT*: https://portswigger.net/blog/goodbye-state-files-we-wont-miss-you",
        "",
        "Returns fuzz results' from a Burp saved state file. This payload's code is based on burp2xml.py:",
        "Developed by Paul Haas, <phaas AT redspin DOT com> under Redspin. Inc.",
        "Licensed under the GNU Public License version 3.0 (2008-2009)",
        "Process Burp Suite Professional's output into a well-formed XML document.",
        "",
        "Currently, the whole burp state file is read, in the future this needs to be changed to be more memory efficient.",
    )
    summary = "Returns fuzz results from a Burp state."
    category = ["default"]
    priority = 99

    parameters = (
        ("fn", "", True, "Filename of a valid Burp state file."),
        ("attr", None, False, "Fuzzresult attribute to return. If not specified the whole object is returned."),
        ("source", "proxy, target", False, "A list of separated Burp sources to get the HTTP requests and responses from. It could be proxy or target tool."),
        ("checkversion", False, False, "If the Burp log file version is unknown an exception will be raised and execution will fail. Checked with burp state file version 65, 67."),
    )

    default_parameter = "fn"

    def __init__(self, params):
        BasePayload.__init__(self, params)

        self.__max = -1
        self.attr = self.params["attr"]
        self._it = self.burp_to_xml(self.params["fn"])

        if any(i not in ["proxy", "target"] for i in self.params["source"].split(",")):
            raise FuzzExceptBadOptions("Unknown burp source parameter")

        self.request_tags = []
        self.response_tags = []

        if "proxy" in self.params["source"]:
            self.request_tags.append("</originalRequest>")
            self.response_tags.append("</originalResponse>")

        if "target" in self.params["source"]:
            self.request_tags.append("</request>")
            self.response_tags.append("</response>")

    def __iter__(self):
        return self

    def count(self):
        return self.__max

    def __next__(self):
        next_item = next(self._it)

        return next_item if not self.attr else rgetattr(next_item, self.attr)

    def milliseconds_to_date(self, milliseconds):
        '''Convert milliseconds since Epoch (from Java) to Python date structure:
        See: http://java.sun.com/j2se/1.4.2/docs/api/java/util/Date.html

        There is no direct way to convert milliseconds since Epoch to Python object
        So we convert the milliseconds to seconds first as a POSIX timestamp which
        can be used to get a valid date, and then use the parsed values from that
        object along with converting mili -> micro seconds in a new date object.'''
        try:
            d = datetime.datetime.fromtimestamp(milliseconds / 1000)
            date = datetime.datetime(d.year, d.month, d.day, d.hour, d.minute, d.second, (milliseconds % 1000) * 1000)
        except ValueError:  # Bad date, just return the milliseconds
            date = str(milliseconds)
            return None
        return date

    def burp_binary_field(self, field, i):
        '''Strip Burp Suite's binary format characters types from our data.
        The first character after the leading tag describes the type of the data.'''
        if len(field) <= i:
            return None, -1
        elif field[i] == '\x00':  # 4 byte integer value
            return str(struct.unpack('>I', field[i + 1:i + 5])[0]), 5
        elif field[i] == '\x01':  # Two possible unsigned long long types
            if field[i + 1] == '\x00':  # (64bit) 8 Byte Java Date
                ms = struct.unpack('>Q', field[i + 1:i + 9])[0]
                date = self.milliseconds_to_date(ms)
                value = date.ctime() if date else 0  # Use the ctime string format for date
            else:  # Serial Number only used ocasionally in Burp
                value = str(struct.unpack('>Q', field[i + 1:i + 9])[0])
            return value, 9
        elif field[i] == '\x02':  # Boolean Object True/False
            return str(struct.unpack('?', field[i + 1:i + 2])[0]), 2
        elif field[i] == '\x03' or field[i] == '\x04':  # 4 byte length + string
            length = struct.unpack('>I', field[i + 1:i + 5])[0]
            # print "Saw string of length", length, "at", i + 5, i + 5+length
            value = field[i + 5:i + 5 + length]
            if '<' in value or '>' in value or '&' in value:  # Sanatize HTML w/CDATA
                value = '<![CDATA[' + value.replace(']]>', ']]><![CDATA[') + ']]>'
            value = ''.join(c for c in value if c in nvprint)  # Remove nonprintables
            return value, 5 + length  # ** TODO: Verify length by matching end tag **
        print("Unknown binary format", repr(field[i]))
        return None, -1

    def strip_cdata(self, data):
        if data.startswith('<![CDATA['):
            data = data[9:]

        if data.endswith(']]>'):
            data = data[:-3]

        return data

    def burp_to_xml(self, filename):
        '''Unzip Burp's file, remove non-printable characters, CDATA any HTML,
        include a valid XML header and trailer, and return a valid XML string.'''

        z = zipfile.ZipFile(self.find_file(filename))  # Open Burp's zip file
        burp = z.read('burp', 'rb')  # Read-in the main burp file
        m = TAG.match(burp, 0)  # Match a tag at the start of the string
        while m:
            index = m.end()
            etag = m.group().replace('<', '</')  # Matching tag

            m = TAG.match(burp, index)  # Attempt to get the next tag
            if not m:  # Data folows
                # Read the type of data using Burp's binary data headers
                value, length = self.burp_binary_field(burp, index)
                if value is None:
                    break

                index += length + len(etag)  # Point our index to the next tag
                m = TAG.match(burp, index)  # And retrieve it

                if self.params["checkversion"] and etag == "</version>" and value not in ["65", "67"]:
                    raise FuzzExceptBadFile("Unknown burp log version %s" % value)

                if etag == "</https>":
                    https_tag = value == "True"

                if etag in self.request_tags:
                    raw_request = self.strip_cdata(value)

                if etag in self.response_tags:
                    fr = FuzzRequest()
                    fr.update_from_raw_http(raw_request, "http" if not https_tag else "https", self.strip_cdata(value))
                    frr = FuzzResult(history=fr)

                    raw_request = ""
                    https_tag = ""

                    yield frr.update()



from .plugin_api.urlutils import parse_url
from .exception import FuzzExceptBadInstall

# python 2 and 3
import sys
if sys.version_info >= (3, 0):
    from urllib.parse import urljoin
else:
    from urlparse import urljoin


class FuzzRequestSoupMixing(object):
    def get_soup(self):
        try:
            from bs4 import BeautifulSoup
        except ImportError:
            raise FuzzExceptBadInstall("You need to install beautifulsoup4 first!")

        soup = BeautifulSoup(self.content, 'html.parser')

        return soup


class FuzzRequestUrlMixing(object):
    # urlparse functions
    @property
    def urlparse(self):
        return parse_url(self.url)

    @property
    def urlp(self):
        return parse_url(self.url)

    @property
    def pstrip(self):
        return self.to_cache_key()

    @property
    def is_path(self):
        if self.code == 200 and self.url[-1] == '/':
            return True
        elif self.code >= 300 and self.code < 400:
            if "Location" in self.headers.response and self.headers.response["Location"][-1] == '/':
                return True
        elif self.code == 401:
            if self.url[-1] == '/':
                return True

        return False

    @property
    def recursive_url(self):
        if self.code >= 300 and self.code < 400 and "Location" in self.headers.response:
            new_url = self.headers.response["Location"]
            if not new_url[-1] == '/':
                new_url += "/"
            # taking into consideration redirections to /xxx/ without full URL
            new_url = urljoin(self.url, new_url)
        elif self.code == 401 or self.code == 200:
            new_url = self.url
            if not self.url[-1] == '/':
                new_url = "/"
        else:
            raise Exception("Error generating recursive url")

        return new_url + "FUZZ"



import re

# Python 2 and 3
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

from wfuzz.plugin_api.mixins import DiscoveryPluginMixin
from wfuzz.plugin_api.base import BasePlugin
from wfuzz.plugin_api.urlutils import parse_url
from wfuzz.externals.moduleman.plugin import moduleman_plugin


@moduleman_plugin
class links(BasePlugin, DiscoveryPluginMixin):
    name = "links"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Parses HTML looking for new content."
    description = ("Parses HTML looking for new content",)
    category = ["active", "discovery"]
    priority = 99

    parameters = (
    )

    def __init__(self):
        BasePlugin.__init__(self)

        regex = [
            r'href="((?!mailto:|tel:|#|javascript:).*?)"',
            r'src="((?!javascript:).*?)"',
            r'action="((?!javascript:).*?)"',
            # http://en.wikipedia.org/wiki/Meta_refresh
            r'<meta.*content="\d+;url=(.*?)">',
            r'getJSON\("(.*?)"',
        ]

        self.regex = []
        for i in regex:
            self.regex.append(re.compile(i, re.MULTILINE | re.DOTALL))

    def validate(self, fuzzresult):
        return fuzzresult.code in [200]

    def process(self, fuzzresult):
        list_links = []

        # <a href="www.owasp.org/index.php/OWASP_EU_Summit_2008">O
        # ParseResult(scheme='', netloc='', path='www.owasp.org/index.php/OWASP_EU_Summit_2008', params='', query='', fragment='')

        for r in self.regex:
            for i in r.findall(fuzzresult.history.content):
                parsed_link = parse_url(i)

                if (not parsed_link.scheme or parsed_link.scheme == "http" or parsed_link.scheme == "https") and (not parsed_link.netloc and parsed_link.path):
                    if i not in list_links:
                        list_links.append(i)

                        # dir path
                        split_path = parsed_link.path.split("/")
                        newpath = '/'.join(split_path[:-1]) + "/"
                        self.queue_url(urljoin(fuzzresult.url, newpath))

                        # file path
                        self.queue_url(urljoin(fuzzresult.url, i))



import socket
import csv as csvmod
import json as jjson
from xml.dom import minidom

from wfuzz.externals.moduleman.plugin import moduleman_plugin
from wfuzz.plugin_api.base import BasePrinter
from wfuzz.fuzzobjects import FuzzResult


@moduleman_plugin
class magictree(BasePrinter):
    name = "magictree"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Prints results in magictree format"
    category = ["default"]
    priority = 99

    def __init__(self, output):
        BasePrinter.__init__(self, output)
        self.node_mt = None
        self.node_service = None

    def __create_xml_element(self, parent, caption, text):
        # Create a <xxx> element
        doc = minidom.Document()
        el = doc.createElement(caption)
        parent.appendChild(el)

        # Give the <xxx> element some text
        ptext = doc.createTextNode(text)

        el.appendChild(ptext)
        return el

    def header(self, summary):
        doc = minidom.Document()

        # <magictree class="MtBranchObject">
        self.node_mt = doc.createElement("magictree")
        self.node_mt.setAttribute("class", "MtBranchObject")

        # <testdata class="MtBranchObject">
        node_td = doc.createElement("testdata")
        node_td.setAttribute("class", "MtBranchObject")
        self.node_mt.appendChild(node_td)

        # <host>209.85.146.105
        host = summary.seed.history.host
        if host.find(":") > 0:
            host, port = host.split(":")
        else:
            port = 80
            if summary.seed.history.scheme.lower() == "https":
                port = 443

        try:
            resolving = socket.gethostbyname(host)
            node_h = self.__create_xml_element(node_td, "host", str(resolving))
        except socket.gaierror:
            node_h = self.__create_xml_element(node_td, "host", str(host))

        # <ipproto>tcp
        node_ipr = self.__create_xml_element(node_h, "ipproto", "tcp")

        # <port>80<state>open</state><service>http
        node_port = self.__create_xml_element(node_ipr, "port", str(port))
        self.__create_xml_element(node_port, "state", "open")
        if summary.seed.history.scheme.lower() == "https":
            node_port = self.__create_xml_element(node_port, "tunnel", "ssl")

        self.node_service = self.__create_xml_element(node_port, "service", "http")

    def result(self, fuzz_result):
        if fuzz_result.type == FuzzResult.result:
            node_url = self.__create_xml_element(self.node_service, "url", str(fuzz_result.url))

            if 'Server' in fuzz_result.history.headers.response:
                self.__create_xml_element(node_url, "HTTPServer", fuzz_result.history.headers.response['Server'])

            location = ""
            if 'Location' in fuzz_result.history.headers.response:
                location = fuzz_result.history.headers.response['Location']

            if fuzz_result.code == 301 or fuzz_result.code == 302 and location:
                self.__create_xml_element(node_url, "RedirectLocation", location)

            self.__create_xml_element(node_url, "ResponseCode", str(fuzz_result.code))
            self.__create_xml_element(node_url, "source", "WFuzz")

    def footer(self, summary):
        self.f.write(self.node_mt.toxml())


@moduleman_plugin
class html(BasePrinter):
    name = "html"
    author = ("Carlos del Ojo", "Christian Martorella", "Adapted to newer versions Xavi Mendez (@xmendez)")
    version = "0.1"
    summary = "Prints results in html format"
    category = ["default"]
    priority = 99

    def __init__(self, output):
        BasePrinter.__init__(self, output)

    def header(self, summary):
        url = summary.url

        self.f.write("<html><head></head><body bgcolor=#000000 text=#FFFFFF><h1>Fuzzing %s</h1>\r\n<table border=\"1\">\r\n<tr><td>#request</td><td>Code</td><td>#lines</td><td>#words</td><td>Url</td></tr>\r\n" % (url))

    def result(self, fuzz_result):
        if fuzz_result.type == FuzzResult.result:
            htmlc = "<font>"

            if fuzz_result.code >= 400 and fuzz_result.code < 500:
                htmlc = "<font color=#FF0000>"
            elif fuzz_result.code >= 300 and fuzz_result.code < 400:
                htmlc = "<font color=#8888FF>"
            elif fuzz_result.code >= 200 and fuzz_result.code < 300:
                htmlc = "<font color=#00aa00>"

            if fuzz_result.history.method.lower() == "post":
                inputs = ""
                for n, v in list(fuzz_result.history.params.post.items()):
                    inputs += "<input type=\"hidden\" name=\"%s\" value=\"%s\">" % (n, v)

                self.f.write("\r\n<tr><td>%05d</td>\r\n<td>%s%d</font></td>\r\n<td>%4dL</td>\r\n<td>%5dW</td>\r\n<td><table><tr><td>%s</td><td><form method=\"post\" action=\"%s\">%s<input type=submit name=b value=\"send POST\"></form></td></tr></table></td>\r\n</tr>\r\n" % (fuzz_result.nres, htmlc, fuzz_result.code, fuzz_result.lines, fuzz_result.words, fuzz_result.description, fuzz_result.url, inputs))
            else:
                self.f.write("\r\n<tr><td>%05d</td><td>%s%d</font></td><td>%4dL</td><td>%5dW</td><td><a href=%s>%s</a></td></tr>\r\n" % (fuzz_result.nres, htmlc, fuzz_result.code, fuzz_result.lines, fuzz_result.words, fuzz_result.url, fuzz_result.url))

    def footer(self, summary):
        self.f.write("</table></body></html><h5>Wfuzz by EdgeSecurity<h5>\r\n")


@moduleman_plugin
class json(BasePrinter):
    name = "json"
    summary = "Results in json format"
    author = ("Federico (@misterade)", "Minor rework by Ilya Glotov (@ilyaglow)")
    version = "0.2"
    category = ["default"]
    priority = 99

    def __init__(self, output):
        BasePrinter.__init__(self, output)
        self.json_res = []

    def header(self, res):
        pass

    def result(self, res):
        if res.type == FuzzResult.result:
            server = ""
            if 'Server' in res.history.headers.response:
                server = res.history.headers.response['Server']
            location = ""
            if 'Location' in res.history.headers.response:
                location = res.history.headers.response['Location']
            elif res.history.url != res.history.redirect_url:
                location = "(*) %s" % res.history.url
            post_data = []
            if res.history.method.lower() == "post":
                for n, v in list(res.history.params.post.items()):
                    post_data.append({"parameter": n, "value": v})

            res_entry = {
                "chars": res.chars,
                "code": res.code,
                "payload": res.description,
                "lines": res.lines,
                "location": location,
                "method": res.history.method,
                "post_data": post_data,
                "server": server,
                "url": res.url,
                "words": res.words
            }
            self.json_res.append(res_entry)

    def footer(self, summary):
        self.f.write(jjson.dumps(self.json_res))


@moduleman_plugin
class raw(BasePrinter):
    name = "raw"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Raw output format"
    category = ["default"]
    priority = 99

    def __init__(self, output):
        BasePrinter.__init__(self, output)

    def header(self, summary):
        self.f.write("Target: %s\n" % summary.url)

        if summary.total_req > 0:
            self.f.write("Total requests: %d\n" % summary.total_req)
        else:
            self.f.write("Total requests: <<unknown>>\n")

        if self.verbose:
            self.f.write("==============================================================================================================================================\n")
            self.f.write("ID    C.Time   Response   Lines      Word         Chars                  Server                                             Redirect   Payload    \n")
            self.f.write("==============================================================================================================================================\n")
        else:
            self.f.write("==================================================================\n")
            self.f.write("ID    Response   Lines      Word         Chars          Request    \n")
            self.f.write("==================================================================\n")

    def _print_verbose(self, res):
        self.f.write("%05d:  " % res.nres)
        self.f.write("%.3fs   C=" % res.timer)

        location = ""
        if 'Location' in res.history.headers.response:
            location = res.history.headers.response['Location']
        elif res.history.url != res.history.redirect_url:
            location = "(*) %s" % res.history.url

        server = ""
        if 'Server' in res.history.headers.response:
            server = res.history.headers.response['Server']

        if res.exception:
            self.f.write("XXX")
        else:
            self.f.write("%05d:  C=%03d" % (res.nres, res.code))

        self.f.write("   %4d L\t   %5d W\t  %5d Ch  %20.20s  %51.51s   \"%s\"\n" % (res.lines, res.words, res.chars, server[:17], location[:48], res.description))

        for i in res.plugins_res:
            self.f.write("  |_ %s\n" % i.issue)

    def _print(self, res):
        if res.exception:
            self.f.write("XXX")
        else:
            self.f.write("%05d:  C=%03d" % (res.nres, res.code))

        self.f.write("   %4d L\t   %5d W\t  %5d Ch\t  \"%s\"\n" % (res.lines, res.words, res.chars, res.description))

        for i in res.plugins_res:
            self.f.write("  |_ %s\n" % i.issue)

    def result(self, res):
        if res.type == FuzzResult.result:
            if self.verbose:
                self._print_verbose(res)
            else:
                self._print(res)

    def footer(self, summary):
        self.f.write("\n")
        self.f.write("Total time: %s\n" % str(summary.totaltime)[:8])

        if summary.backfeed() > 0:
            self.f.write("Processed Requests: %s (%d + %d)\n" % (str(summary.processed())[:8], (summary.processed() - summary.backfeed()), summary.backfeed()))
        else:
            self.f.write("Processed Requests: %s\n" % (str(summary.processed())[:8]))
        self.f.write("Filtered Requests: %s\n" % (str(summary.filtered())[:8]))
        self.f.write("Requests/sec.: %s\n" % str(summary.processed() / summary.totaltime if summary.totaltime > 0 else 0)[:8])


@moduleman_plugin
class csv(BasePrinter):
    name = "csv"
    author = ("@Yoginski initial version", "Adapted by @egilas to work in newer version of wfuzz")
    summary = "CSV printer ftw"
    version = "1.0"
    category = ["default"]
    priority = 99

    def write(self, e):
        self.f.write(e)
        pass

    def __init__(self, output):
        BasePrinter.__init__(self, output)
        self.csv_writer = csvmod.writer(self)

    def header(self, summary):
        self._print_csv(["id", "response", "lines", "word", "chars", "request", "success"])

    def result(self, res):
        if res.type == FuzzResult.result:
            line = [res.nres,
                    res.code,
                    res.lines,
                    res.words,
                    res.chars,
                    res.description,
                    0 if res.exception else 1]
            self._print_csv(line)

    def noresult(self, res):
        pass

    def footer(self, summary):
        pass

    def _print_csv(self, values):
        self.csv_writer.writerow(values)



import os


# Python 2 and 3
import sys
if sys.version_info >= (3, 0):
    from urllib.parse import ParseResult
    from urllib.parse import urlparse
else:
    from urlparse import ParseResult
    from urlparse import urlparse

from wfuzz.facade import Facade
from wfuzz.exception import FuzzExceptBadAPI


class FuzzRequestParse(ParseResult):
    @property
    def ffname(self):
        '''
        Returns script plus extension from an URL. ie. http://www.localhost.com/kk/index.html?id=3
        will return index.html
        '''
        u = self.path.split('/')[-1:][0]

        return u

    @property
    def fext(self):
        '''
        Returns script extension from an URL. ie. http://www.localhost.com/kk/index.html?id=3
        will return .html
        '''
        return os.path.splitext(self.ffname)[1]

    @property
    def fname(self):
        '''
        Returns script name from an URL. ie. http://www.localhost.com/kk/index.html?id=3
        will return index
        '''
        return os.path.splitext(self.ffname)[0]

    @property
    def isbllist(self):
        fext = self.fext
        return fext != "." and fext in Facade().sett.get("kbase", "discovery.blacklist").split("-")

    @property
    def hasquery(self):
        return self.query != ""


def parse_url(url):
    scheme, netloc, path, params, query, fragment = urlparse(url)
    return FuzzRequestParse(scheme, netloc, path, params, query, fragment)


def check_content_type(fuzzresult, which):
    ctype = None
    if 'Content-Type' in fuzzresult.history.headers.response:
        ctype = fuzzresult.history.headers.response['Content-Type']

    if which == 'text':
        return not ctype or (ctype and any([ctype.find(x) >= 0 for x in ['text/plain']]))
    else:
        raise FuzzExceptBadAPI("Unknown content type")
